@@ -31,16 +31,22 @@ import signal
 import subprocess
 import tempfile
 import shutil
+from string import Template
 import urllib.parse
-from typing import Set
+from typing import Optional, Set
+
+from debian.debian_support import Version
+from debian.deb822 import Deb822
 
 import adtlog
 import VirtSubproc
+from VirtSubproc import load_shell_script
 from testdesc import Unsupported
 
 
 timeouts = {'short': 100, 'copy': 300, 'install': 3000, 'test': 10000,
             'build': 100000}
+global_timeout = 0
 
 
 # When running installed, this is /usr/share/autopkgtest.
@@ -84,6 +90,7 @@ class Testbed:
         self.eatmydata_prefix = []
         self.apt_pin_for_releases = []
         self.enable_apt_fallback = enable_apt_fallback
+        self.apt_version: Optional[Version] = None
         self.needs_internet = needs_internet
         self.shell_fail = shell_fail
         self.nproc = None
@@ -94,6 +101,11 @@ class Testbed:
         # False if not tried, True if successful, error message if not
         self._tried_provide_sudo = False
         self._guessed_default_release = None
+        # used for keeping track of the global timeout, if any
+        self.start_time = int(time.time())
+        self.skip_remaining_tests = False
+        # Absolute path to wrapper.sh on testbed
+        self.wrapper_sh = ''
 
         try:
             self.devnull = subprocess.DEVNULL
@@ -169,7 +181,7 @@ class Testbed:
         # usually "noexec" and /[s]bin might be readonly, so create in /tmp
         if 'reboot' in self.caps and 'root-on-testbed' in self.caps:
             adtlog.debug('testbed supports reboot, creating /tmp/autopkgtest-reboot')
-            reboot = Path(
+            reboot = TestbedPath(
                 self,
                 os.path.join(PKGDATADIR, 'lib', 'in-testbed', 'reboot.sh'),
                 '%s/autopkgtest-reboot' % self.scratch,
@@ -189,7 +201,7 @@ class Testbed:
                 stderr=subprocess.PIPE,
             )
 
-            reboot_prepare = Path(
+            reboot_prepare = TestbedPath(
                 self,
                 os.path.join(PKGDATADIR, 'lib', 'in-testbed',
                              'reboot-prepare.sh'),
@@ -260,11 +272,13 @@ class Testbed:
             if c.startswith('downtmp-host='):
                 self.shared_downtmp = c.split('=', 1)[1]
 
-        Path(
+        wrapper_sh = TestbedPath(
             self, os.path.join(PKGDATADIR, 'lib', 'in-testbed', 'wrapper.sh'),
             '%s/wrapper.sh' % self.scratch,
             False,
-        ).copydown(mode='0755')
+        )
+        wrapper_sh.copydown(mode='0755')
+        self.wrapper_sh = wrapper_sh.tb
 
         # provide a default for --user
         if self.user is None and 'root-on-testbed' in self.caps:
@@ -297,12 +311,28 @@ class Testbed:
             ) as reader:
                 self.check_exec(['sh', '-euc', reader.read(), 'sh', self.user])
 
-        self.run_setup_commands()
-
         # determine testbed architecture
         self.dpkg_arch = self.check_exec(['dpkg', '--print-architecture'], True).strip()
         adtlog.info('testbed dpkg architecture: ' + self.dpkg_arch)
 
+        have_dpkg_query = self.execute(
+            ['sh', '-ec', 'command -v dpkg-query'],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+        )[0] == 0
+
+        if have_dpkg_query:
+            out = self.check_exec(['dpkg-query', '-W', '-f', '${Version}', 'apt'], True).strip()
+            adtlog.info('testbed apt version: ' + out)
+            try:
+                self.apt_version = Version(out)
+            except ValueError:
+                self.apt_version = None
+        else:
+            self.apt_version = None
+
+        self.run_setup_commands()
+
         # do we have eatmydata?
         (code, out, err) = self.execute(
             ['sh', '-ec', 'command -v eatmydata'],
@@ -313,12 +343,16 @@ class Testbed:
             adtlog.debug('testbed has eatmydata')
             self.eatmydata_prefix = [out.strip()]
 
-        # record package versions of pristine testbed
-        if self.output_dir and self.execute(
+        # TODO: UGLY HACK for now because test_no_dpkg_query fails
+        # otherwise
+        have_dpkg_query = self.execute(
             ['sh', '-ec', 'command -v dpkg-query'],
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE,
-        )[0] == 0:
+        )[0] == 0
+
+        # record package versions of pristine testbed
+        if self.output_dir and have_dpkg_query:
             pkglist = TempPath(self, 'testbed-packages', autoclean=False)
             self.check_exec(['sh', '-ec', "dpkg-query --show -f '${Package}\\t${Version}\\n' > %s" % pkglist.tb])
             pkglist.copyup()
@@ -359,7 +393,7 @@ class Testbed:
         adtlog.info('@@@@@@@@@@@@@@@@@@@@ test bed setup')
         for (host, tb) in self.copy_files:
             adtlog.debug('Copying file %s to testbed %s' % (host, tb))
-            Path(self, host, tb, os.path.isdir(host)).copydown()
+            TestbedPath(self, host, tb, os.path.isdir(host)).copydown()
 
         # Make sure default release is set regardless of other options if so
         # required by --apt-default-release. Typically this would be used in
@@ -372,25 +406,12 @@ class Testbed:
             self._add_apt_source(source)
 
         # create apt sources for --apt-pocket
+        if self.add_apt_pockets:
+            script = load_shell_script(
+                'setup-commands/create-sources.list-for-pocket')
         for pocket in self.add_apt_pockets:
             pocket = pocket.split('=', 1)[0]  # strip off package list
-            script = '''
-            sed -rn 's/^(deb|deb-src) +(\\[.*\\] *)?((http|https|file):[^ ]*) +([^ -]+) +(.*)$/\\1 \\2\\3 \\5-%s \\6/p' \\
-                /etc/apt/sources.list `ls /etc/apt/sources.list.d/*.list 2>/dev/null|| true` > \\
-                /etc/apt/sources.list.d/%s.list
-            for retry in 1 2 3; do
-                apt-get --no-list-cleanup \\
-                        -o Dir::Etc::sourcelist=/etc/apt/sources.list.d/%s.list \\
-                        -o Dir::Etc::sourceparts=/dev/null \\
-                        update 2>&1 && break || \\
-                        if [ $retry = 3 ] ; then
-                            return 1
-                        else
-                            sleep 15
-                        fi
-            done
-            ''' % (pocket, pocket, pocket)
-            self.check_exec(['sh', '-ec', script])
+            self.check_exec(['sh', '-euc', script, 'sh', pocket])
 
         # create apt pinning for --apt-pocket
         for pocket in self.add_apt_pockets:
@@ -516,13 +537,13 @@ class Testbed:
             # Also check whether they need a password (by default yes,
             # but in e.g. Ubuntu cloud images they don't)
             rc, out, err = self.execute(
-                run_as_user + ['sudo -n -v'],
+                run_as_user + ['sudo -n true'],
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
             )
-            adtlog.debug('sudo -n -v stdout: %s' % out.strip())
-            adtlog.debug('sudo -n -v stderr: %s' % err.strip())
-            adtlog.debug('sudo -n -v status: %d' % rc)
+            adtlog.debug('sudo -n true stdout: %s' % out.strip())
+            adtlog.debug('sudo -n true stderr: %s' % err.strip())
+            adtlog.debug('sudo -n true status: %d' % rc)
             already_has_sudo = (rc == 0)
         else:
             adtlog.debug('id -Gn stdout: %s' % out.strip())
@@ -688,7 +709,20 @@ class Testbed:
         if env:
             argv = ['env'] + env + argv
 
-        VirtSubproc.timeout_start(timeouts[kind])
+        abort_after_timeout = False
+        timeout = timeouts[kind]
+        timeout_type = 'timed out'
+        if kind in ['test', 'build'] and global_timeout > 0:
+            remaining_time = self.start_time + global_timeout - int(time.time())
+            if remaining_time <= 0:
+                adtlog.error('global timeout reached even before %s' % ' '.join(argv))
+                raise VirtSubproc.Timeout(global_timeout, abort=True)
+            if timeout > remaining_time:
+                timeout = remaining_time
+                timeout_type = 'global timeout reached'
+                abort_after_timeout = True
+
+        VirtSubproc.timeout_start(timeout)
         try:
             proc = subprocess.Popen(self.exec_cmd + argv,
                                     stdin=self.devnull,
@@ -699,18 +733,23 @@ class Testbed:
             if err is not None:
                 err = err.decode()
             VirtSubproc.timeout_stop()
-        except VirtSubproc.Timeout:
+        except VirtSubproc.Timeout as exc:
             # This is a bit of a hack, but what can we do.. we can't kill/clean
             # up sudo processes, we can only hope that they clean up themselves
             # after we stop the testbed
             killtree(proc.pid)
-            adtlog.debug('timed out on %s %s (kind: %s)' % (self.exec_cmd, argv, kind))
+            adtlog.debug('%s on %s %s (kind: %s)' % (timeout_type, self.exec_cmd, argv, kind))
             if 'sudo' not in self.exec_cmd:
-                proc.wait()
-            msg = 'timed out on command "%s" (kind: %s)' % (' '.join(argv), kind)
+                try:
+                    proc.wait(timeout=10)
+                except subprocess.TimeoutExpired:
+                    killtree(proc.pid, signal.SIGKILL)
+                    proc.wait()
+            msg = '%s on command "%s" (kind: %s)' % (timeout_type, ' '.join(argv), kind)
             if kind == 'test':
                 adtlog.error(msg)
-                raise
+                exc.abort = abort_after_timeout
+                raise exc
             else:
                 self.debug_fail()
                 self.bomb(msg)
@@ -823,12 +862,18 @@ Description: satisfy autopkgtest test dependencies
         shutil.rmtree(pkgdir)
         deb.copydown()
 
+        apt_version = self.apt_version
+        assert apt_version is not None
+        # Debian >= 9, Ubuntu >= 16.04
+        apt_can_install_debs = (apt_version >= Version('1.1~exp1'))
+
         # install it and its dependencies in the tb; our apt pinning is not
         # very clever wrt. resolving transitional dependencies in the pocket,
         # so we might need to retry without pinning
         download_fail_retries = 3
         while True:
-            self.check_exec(['dpkg', '--unpack', deb.tb], stdout=subprocess.PIPE)
+            if not apt_can_install_debs:
+                self.check_exec(['dpkg', '--unpack', deb.tb], stdout=subprocess.PIPE)
 
             # if '@' was provided as a test dependency (as seen in synth_deps),
             # we explicitly install the corresponding binary packages in case
@@ -836,7 +881,11 @@ Description: satisfy autopkgtest test dependencies
 
             rc = 0
             try:
-                self._run_apt_install('--fix-broken', ' '.join(self.eatmydata_prefix))
+                if apt_can_install_debs:
+                    self._run_apt_install(deb.tb, ' '.join(self.eatmydata_prefix))
+                else:
+                    self._run_apt_install('--fix-broken', ' '.join(self.eatmydata_prefix))
+
                 need_explicit_install = []
                 for dep in synth_deps:
                     if dep != list(dep):
@@ -887,8 +936,7 @@ Description: satisfy autopkgtest test dependencies
                 if shell_on_failure:
                     self.run_shell()
             else:
-                # apt-get -f may succeed, but its solution might remove
-                # autopkgtest-satdep, which is still a failure
+                # verify that we actually got autopkgtest-satdep installed
                 rc = self.execute(['dpkg', '--status', 'autopkgtest-satdep'],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)[0]
@@ -963,30 +1011,23 @@ Description: satisfy autopkgtest test dependencies
 
         # resolve arch specific dependencies; don't use universal_newlines
         # here, it's broken for stdin on Python 3.2
-        if build_dep:
-            extra_args = ', reduce_profiles => $supports_profiles, build_dep => 1'
-        else:
-            extra_args = ''
-        perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
-                                stdout=subprocess.PIPE)
-        code = '''use Dpkg::Deps;
-                  $supports_profiles = ($Dpkg::Deps::VERSION gt '1.04' or 0);
-                  $origdeps = '%s';
-                  $origdeps =~ s/(^|,)[^<,]+<[^!,>]+>//g if (!$supports_profiles);
-                  $dep = deps_parse($origdeps, reduce_arch => 1, host_arch => '%s' %s);
-                  $out = $dep->output();
-                  # fall back to ignoring build profiles
-                  $out =~ s/ <![^ >]+>//g if (!$supports_profiles);
-                  print $out, "\\n";
-                  ''' % (deps, self.dpkg_arch, extra_args)
-        deps = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
-        if perl.returncode != 0:
-            self.bomb('failed to run perl for parsing dependencies')
+        subproc = subprocess.Popen(
+            [
+                os.path.join(PKGDATADIR, 'lib', 'parse-deps.pl'),
+                deps,
+                self.dpkg_arch,
+                'build-depends' if build_dep else 'depends',
+            ],
+            stdout=subprocess.PIPE,
+        )
+        deps = subproc.communicate()[0].decode('UTF-8').strip()
+        if subproc.returncode != 0:
+            self.bomb(f'failed to run {subproc.args[0]}')
         adtlog.debug('%s: architecture resolved: %s' % (what, deps))
 
         # check if we can use apt-get
         can_apt_get = False
-        if 'root-on-testbed' in self.caps:
+        if 'root-on-testbed' in self.caps and self.apt_version is not None:
             rc = self.execute(['test', '-w', '/var/lib/dpkg/status'])[0]
             if rc == 0:
                 can_apt_get = True
@@ -1022,13 +1063,17 @@ Description: satisfy autopkgtest test dependencies
                  shell=False, build_parallel=None):
         '''Run given test in testbed
 
-        tree (a Path) is the source tree root.
+        tree (a TestbedPath) is the source tree root.
         '''
         def _info(m):
             adtlog.info('test %s: %s' % (test.name, m))
 
         self.last_test_name = test.name
 
+        if self.skip_remaining_tests:
+            test.set_skipped('global timeout exceeded')
+            return
+
         if test.path and not os.path.exists(os.path.join(tree.host, test.path)):
             self.badpkg('%s does not exist' % test.path)
 
@@ -1064,50 +1109,61 @@ Description: satisfy autopkgtest test dependencies
         test_artifacts = '%s/%s-artifacts' % (self.scratch, test.name)
         autopkgtest_tmp = '%s/autopkgtest_tmp' % (self.scratch)
         assert self.nproc is not None
-        # TODO: Move more of this into wrapper.sh
-        script = 'set -e; ' \
-                 'export USER=`id -nu`; ' \
-                 '. /etc/profile >/dev/null 2>&1 || true; ' \
-                 ' . ~/.profile >/dev/null 2>&1 || true; ' \
-                 'buildtree="%(t)s"; ' \
-                 'mkdir -p -m 1777 -- "%(a)s"; ' \
-                 'export AUTOPKGTEST_ARTIFACTS="%(a)s"; ' \
-                 'export ADT_ARTIFACTS="$AUTOPKGTEST_ARTIFACTS"; ' \
-                 'mkdir -p -m 755 "%(tmp)s"; export AUTOPKGTEST_TMP="%(tmp)s"; ' \
-                 'export ADTTMP="$AUTOPKGTEST_TMP"; ' \
-                 'export DEBIAN_FRONTEND=noninteractive; ' \
-                 'export LANG=C.UTF-8; ' \
-                 '''export DEB_BUILD_OPTIONS=parallel=%(cpu)s; ''' \
-                 'unset LANGUAGE LC_CTYPE LC_NUMERIC LC_TIME LC_COLLATE '\
-                 '  LC_MONETARY LC_MESSAGES LC_PAPER LC_NAME LC_ADDRESS '\
-                 '  LC_TELEPHONE LC_MEASUREMENT LC_IDENTIFICATION LC_ALL;' \
-                 'cd "$buildtree"; '\
-                 % {'t': tree.tb, 'a': test_artifacts, 'tmp': autopkgtest_tmp,
-                    'cpu': build_parallel or self.nproc}
+
+        argv = [self.wrapper_sh]
+
+        if adtlog.verbosity >= 2:
+            # This should come first so that it affects all CLI options
+            argv.append('--debug')
+
+        argv.extend([
+            '--artifacts={}'.format(test_artifacts),
+            '--chdir={}'.format(tree.tb),
+            '--env=DEB_BUILD_OPTIONS=parallel={}'.format(
+                build_parallel or self.nproc
+            ),
+            '--env=DEBIAN_FRONTEND=noninteractive',
+            '--env=LANG=C.UTF-8',
+            '--unset-env=LANGUAGE',
+            '--unset-env=LC_ADDRESS',
+            '--unset-env=LC_ALL',
+            '--unset-env=LC_COLLATE',
+            '--unset-env=LC_CTYPE',
+            '--unset-env=LC_IDENTIFICATION',
+            '--unset-env=LC_MEASUREMENT',
+            '--unset-env=LC_MESSAGES',
+            '--unset-env=LC_MONETARY',
+            '--unset-env=LC_NAME',
+            '--unset-env=LC_NUMERIC',
+            '--unset-env=LC_PAPER',
+            '--unset-env=LC_TELEPHONE',
+            '--unset-env=LC_TIME',
+            '--script-pid-file=/tmp/autopkgtest_script_pid',
+            '--source-profile',
+            '--stderr={}'.format(se.tb),
+            '--stdout={}'.format(so.tb),
+            '--tmp={}'.format(autopkgtest_tmp),
+        ])
 
         if 'needs-root' in test.restrictions and self.user is not None:
-            script += 'export AUTOPKGTEST_NORMAL_USER=%s; ' % self.user
-            script += 'export ADT_NORMAL_USER=%s; ' % self.user
+            argv.append('--env=AUTOPKGTEST_NORMAL_USER={}'.format(self.user))
+            argv.append('--env=ADT_NORMAL_USER={}'.format(self.user))
 
         for e in extra_env:
-            script += 'export \'%s\'; ' % e
+            argv.append('--env={}'.format(e))
 
         if test.path:
             test_cmd = os.path.join(tree.tb, test.path)
-            script += 'chmod +x %s; ' % test_cmd
+            argv.extend([
+                '--make-executable=' + test_cmd,
+                '--',
+                test_cmd,
+            ])
         else:
-            test_cmd = "bash -ec '%s'" % test.command.replace("'", "'\\''")
+            argv.extend(['--', 'bash', '-ec', test.command])
 
-        script += ' '.join([
-            'exec',
-            '{}/wrapper.sh'.format(shlex.quote(self.scratch)),
-            '--script-pid-file=/tmp/autopkgtest_script_pid',
-            '--stderr={}'.format(shlex.quote(se.tb)),
-            '--stdout={}'.format(shlex.quote(so.tb)),
-            '--',
-            test_cmd,
-            ';',
-        ])
+        # Convert the argv into a shell one-line so we can wrap it in su
+        script = 'set -e; exec ' + ' '.join(map(shlex.quote, argv))
 
         if 'needs-root' not in test.restrictions and self.user is not None:
             if 'root-on-testbed' not in self.caps:
@@ -1144,9 +1200,10 @@ Description: satisfy autopkgtest test dependencies
                 script_prefix = ''
             try:
                 rc = self.execute(test_argv + [script_prefix + script], kind='test')[0]
-            except VirtSubproc.Timeout:
+            except VirtSubproc.Timeout as exc:
                 rc = 1
                 timeout = True
+                self.skip_remaining_tests = exc.abort
                 break
 
             # If the shell or su command that runs wrapper.sh exits with
@@ -1257,8 +1314,8 @@ Description: satisfy autopkgtest test dependencies
 
         # copy artifacts to host, if we have --output-dir
         if self.output_dir:
-            ap = Path(self, os.path.join(self.output_dir, 'artifacts'),
-                      test_artifacts, is_dir=True)
+            ap = TestbedPath(self, os.path.join(self.output_dir, 'artifacts'),
+                             test_artifacts, is_dir=True)
             ap.copyup()
             # don't keep an empty artifacts dir around
             if not os.listdir(ap.host):
@@ -1288,38 +1345,66 @@ Description: satisfy autopkgtest test dependencies
         '''Create apt pinning for --apt-pocket=pocket=pkglist and
            --pin-packages=release=pkglist'''
 
-        # sort pkglist into source and binary packages
-        binpkgs = []
-        srcpkgs = []
-        for i in pkglist.split(','):
-            i = i.strip()
-            if i.startswith('src:'):
-                srcpkgs.append(i[4:])
-            else:
-                binpkgs.append(i)
+        pref_template = Template("""Package: $pkgs
+Pin: release a=$release
+Pin-Priority: $prio
 
-        script = 'mkdir -p /etc/apt/preferences.d; '
-        script += 'PKGS="%s"; ' % ' '.join(binpkgs)
+""")
+        apt_version = self.apt_version
+        assert apt_version is not None
+        if apt_version < Version('2.0'):
+            # sort pkglist into source and binary packages
+            binpkgs = []
+            srcpkgs = []
+            for i in pkglist.split(','):
+                i = i.strip()
+                if i.startswith('src:'):
+                    srcpkgs.append(i[4:])
+                else:
+                    binpkgs.append(i)
 
-        # translate src:name entries into binaries of that source
-        if srcpkgs:
-            script += 'PKGS="$PKGS $(apt-cache showsrc %s | ' \
-                '''awk '/^Package-List:/ { show=1; next } (/^ / && show==1) { print $1; next } { show=0 }' |''' \
-                '''sort -u | tr '\\n' ' ')"; ''' % \
-                ' '.join(srcpkgs)
+            # translate src:name entries into binaries of that source
+            if srcpkgs:
+                apt_cache = self.check_exec(
+                    ['apt-cache', 'dumpavail'], True)
+                for paragr in Deb822.iter_paragraphs(apt_cache):
+                    pkg = paragr.get('Package')
+                    src = paragr.get('Source', pkg)
+                    if src in srcpkgs and pkg not in binpkgs:
+                        binpkgs.append(pkg)
+
+            pkglist = " ".join(binpkgs)
+
+        # prefer given packages from source suite, but make sure that
+        # other packages are taken from default suite as much as
+        # possible
+        pref_content = pref_template.substitute(
+            pkgs=pkglist.replace(',', ' '),
+            release=release,
+            prio='995')
+        for default in default_releases:
+            pref_content += pref_template.substitute(
+                pkgs="*",
+                release=default,
+                prio='990')
 
         # filenames shouldn't contain slashes, so convert them when
         # they appear in the release name
         # apt also doesn't like unknown extensions, so replace dot too
         release_file_name = release.replace('/', '-').replace('.', '_')
-        # prefer given packages from series, but make sure that other packages
-        # are taken from default release as much as possible
-        script += 'printf "Package: $PKGS\\nPin: release a=%(release)s\\nPin-Priority: 995\\n" > /etc/apt/preferences.d/autopkgtest-%(release_file_name)s; ' % \
-            {'release': release, 'release_file_name': release_file_name}
-        for default in default_releases:
-            script += 'printf "\nPackage: *\\nPin: release a=%(default)s\\nPin-Priority: 990\\n" >> /etc/apt/preferences.d/autopkgtest-%(release_file_name)s; ' % \
-                {'release_file_name': release_file_name, 'default': default}
-        self.check_exec(['sh', '-ec', script])
+
+        with tempfile.NamedTemporaryFile('w+', encoding='UTF-8') as tmp_pref_file:
+            tmp_pref_file.write(pref_content)
+            tmp_pref_file.flush()
+            pref_file = TestbedPath(
+                self,
+                tmp_pref_file.name,
+                '/etc/apt/preferences.d/autopkgtest-' + release_file_name)
+            pref_file.copydown(mode='0644')
+
+        adtlog.debug('autopkgtest-' + release_file_name + ' content:\n' +
+                     pref_content)
+
         self._set_default_release()
         self.apt_pin_for_releases.append(release)
 
@@ -1337,8 +1422,7 @@ Description: satisfy autopkgtest test dependencies
 
         # Guess the default release by parsing the sources.list files.
         # The first encountered release will be considered the default release.
-        script = 'SRCS=$(ls /etc/apt/sources.list /etc/apt/sources.list.d/*.list 2>/dev/null|| true); '
-        script += r'''sed -rn '/^(deb|deb-src) +(\[.*\] *)?(http|https|file):/ { s/\[.*\] +//; s/^[^ ]+ +[^ ]* +([^ ]+) +.*$/\1/p }' $SRCS | head -n1'''
+        script = load_shell_script('setup-commands/get-default-release')
         self._guessed_default_release = self.check_exec(['sh', '-ec', script], stdout=True).strip()
 
         return self._guessed_default_release
@@ -1360,28 +1444,15 @@ Description: satisfy autopkgtest test dependencies
 
     def _add_apt_source(self, source, filename='autopkgtest-add-apt-sources.list'):
         adtlog.debug('adding APT source: %s' % source)
-        script = '''
-        echo "$1" >> /etc/apt/sources.list.d/$2
-        for retry in 1 2 3; do
-            apt-get --no-list-cleanup \\
-                    -o Dir::Etc::sourcelist=/etc/apt/sources.list.d/$2 \\
-                    -o Dir::Etc::sourceparts=/dev/null \\
-                    update 2>&1 && break || \\
-                    if [ $retry = 3 ] ; then
-                        return 1
-                    else
-                        sleep 15
-                    fi
-        done
-        '''
+        script = load_shell_script('setup-commands/add-apt-sources')
         self.check_exec(['sh', '-ec', script, 'sh', source, filename])
 
 
-class Path:
+class TestbedPath:
     '''Represent a file/dir with a host and a testbed path'''
 
     def __init__(self, testbed, host, tb, is_dir=None):
-        '''Create a Path object.
+        '''Create a TestbedPath object.
 
         The object itself is just a pair of file names, nothing more. They do
         not need to exist until you call copyup() or copydown() on them.
@@ -1449,7 +1520,7 @@ class Path:
             self.testbed.command('copyup', (self.tb, self.host))
 
 
-class TempPath(Path):
+class TempPath(TestbedPath):
     '''Represent a file in the hosts'/testbed's temporary directories
 
     These are only guaranteed to exit within one testbed run.
@@ -1459,7 +1530,7 @@ class TempPath(Path):
     _filename_prefix = 1
 
     def __init__(self, testbed, name, is_dir=False, autoclean=True):
-        '''Create a temporary Path object.
+        '''Create a temporary TestbedPath object.
 
         The object itself is just a pair of file names, nothing more. They do
         not need to exist until you call copyup() or copydown() on them.
@@ -1485,9 +1556,9 @@ class TempPath(Path):
             name = str(self._filename_prefix) + '-' + name
             TempPath._filename_prefix += 1
 
-        Path.__init__(self, testbed, os.path.join(host, name),
-                      os.path.join(testbed.scratch, name),
-                      is_dir)
+        TestbedPath.__init__(self, testbed, os.path.join(host, name),
+                             os.path.join(testbed.scratch, name),
+                             is_dir)
 
     def __del__(self):
         if self.autoclean:
@@ -1515,13 +1586,13 @@ def child_ps(pid):
         return []
 
 
-def killtree(pid):
+def killtree(pid, kill_signal=signal.SIGTERM):
     '''Recursively kill pid and all of its children'''
 
     for child in child_ps(pid):
-        killtree(child)
+        killtree(child, kill_signal)
     try:
-        os.kill(pid, signal.SIGTERM)
+        os.kill(pid, kill_signal)
     except OSError:
         pass
 
